/*******************************************************************************
 * Copyright (c) 2019, 2020 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Apache License, Version 2.0 which is available at
 * https://www.apache.org/licenses/LICENSE-2.0.
 *
 * SPDX-License-Identifier: Apache-2.0
 *******************************************************************************/
package org.fortiss.tooling.common.ui.javafx.lwfxef;

import static java.lang.Math.abs;
import static java.lang.Math.max;
import static java.lang.Math.min;
import static java.util.Objects.requireNonNull;
import static javafx.geometry.Orientation.HORIZONTAL;
import static javafx.geometry.Orientation.VERTICAL;
import static javafx.geometry.VPos.TOP;
import static javafx.scene.layout.GridPane.setHgrow;
import static javafx.scene.layout.GridPane.setVgrow;
import static javafx.scene.layout.Priority.ALWAYS;
import static javafx.scene.paint.Color.ORANGERED;
import static javafx.scene.paint.Color.grayRgb;
import static org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramViewerDefaultTags.LINK_TARGET_ALLOWED_TAG;
import static org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramViewerDefaultTags.LINK_TARGET_DENIED_TAG;
import static org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramViewerDefaultTags.LINK_TARGET_MAYBE_TAG;
import static org.fortiss.tooling.common.ui.javafx.util.GraphicUtils.getFXImage;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.function.Consumer;

import org.fortiss.tooling.common.ui.ToolingCommonUIActivator;
import org.fortiss.tooling.common.ui.javafx.lwfxef.change.Change;
import org.fortiss.tooling.common.ui.javafx.lwfxef.change.ChangeSet;
import org.fortiss.tooling.common.ui.javafx.lwfxef.change.DefaultModelModifier;
import org.fortiss.tooling.common.ui.javafx.lwfxef.controller.IController;
import org.fortiss.tooling.common.ui.javafx.lwfxef.controller.IControllerFactory;
import org.fortiss.tooling.common.ui.javafx.lwfxef.controller.IDragController;
import org.fortiss.tooling.common.ui.javafx.lwfxef.controller.IKeyPressController;
import org.fortiss.tooling.common.ui.javafx.lwfxef.model.IModelFactory;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IAnchorageMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IContentMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IDiagramAnchorageMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IDiagramMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.ILinkMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.MVCBundleTag;
import org.fortiss.tooling.common.ui.javafx.lwfxef.visual.IVisualFactory;

import javafx.event.EventHandler;
import javafx.geometry.BoundingBox;
import javafx.geometry.Bounds;
import javafx.geometry.Orientation;
import javafx.geometry.Point2D;
import javafx.geometry.Rectangle2D;
import javafx.scene.Node;
import javafx.scene.control.ContextMenu;
import javafx.scene.control.Label;
import javafx.scene.control.MenuItem;
import javafx.scene.control.ScrollBar;
import javafx.scene.image.Image;
import javafx.scene.image.ImageView;
import javafx.scene.input.KeyEvent;
import javafx.scene.input.MouseButton;
import javafx.scene.input.MouseEvent;
import javafx.scene.input.ScrollEvent;
import javafx.scene.layout.GridPane;
import javafx.scene.layout.Pane;
import javafx.scene.paint.Color;
import javafx.scene.paint.Paint;
import javafx.scene.shape.Line;
import javafx.scene.shape.Rectangle;
import javafx.scene.text.Text;
import javafx.scene.transform.Affine;
import javafx.scene.transform.Transform;
import javafx.scene.transform.Translate;

/**
 * This class represents the diagram viewer node, which manages the grid
 * background and the content visuals as well as scrolling, zooming, selection,
 * context menus and link creation. Due to the number of features and functions,
 * part of this class delegates to a {@link MVCBundleManager} object used solely
 * for the purpose of source code size reduction.
 */
public class DiagramViewer {
	/** The default dash stroke pattern. */
	private final Double[] DEFAULT_DASH_STROKE_PATTERN = new Double[] {15.0, 5.0};
	/** The bundle manager helper class reference. */
	private final MVCBundleManager viewerManager;
	/** The feature state helper class reference. */
	private final DiagramViewerFeatures features;
	/** The grid pane for the main pane and scrollbars. */
	private final GridPane scrolledPane;
	/** The horizontal scrollbar. */
	private final ScrollBar horizontalScrollbar;
	/** The vertical scrollbar. */
	private final ScrollBar verticalScrollbar;
	/** The group for layered nodes. */
	private final DiagramLayers layers;
	/** The viewer main pane. */
	private final Pane viewerPane;
	/** The help label. */
	private final Label helpLabel = new Label("");
	/** The help display. */
	private Text helpText = null;
	/** The help shadowed background. */
	private final Rectangle helpShadow = new Rectangle(0, 0);
	/** The mouse drag multi-selection feedback rectangle. */
	private final Rectangle mouseDragRectangle = new Rectangle(0, 0);
	/** The initial location where the selection rectangle was started. */
	private Point2D mouseDragRectangleStartLocation = new Point2D(0, 0);
	/** The line used during link creation gesture. */
	private final Line linkLineFeedback = new Line();
	/** The current context menu. */
	private ContextMenu contextMenu = null;
	/** All scaling and translation of the diagram content is done through this transform. */
	private Affine transform = new Affine();

	/** A debug label. */
	private Label debugLabel = new Label("");

	/** Constructor. */
	public DiagramViewer(IModelFactory modelFactory, IVisualFactory visualFactory,
			IControllerFactory controllerFactory,
			Consumer<DiagramViewerSelection> selectionChangedCallback) {
		this(modelFactory, visualFactory, controllerFactory, selectionChangedCallback,
				new DefaultModelModifier());
	}

	/** Constructor. */
	public DiagramViewer(IModelFactory modelFactory, IVisualFactory visualFactory,
			IControllerFactory controllerFactory,
			Consumer<DiagramViewerSelection> selectionChangedCallback,
			Consumer<Change> modelModifier) {
		features = new DiagramViewerFeatures(this);
		layers = new DiagramLayers(this);
		viewerManager = new MVCBundleManager(this, modelFactory, visualFactory, controllerFactory,
				layers, selectionChangedCallback, requireNonNull(modelModifier));

		// selection feedback rectangle
		mouseDragRectangle.setFill(Color.TRANSPARENT);
		mouseDragRectangle.setStroke(Color.ORANGERED);
		mouseDragRectangle.getStrokeDashArray().addAll(DEFAULT_DASH_STROKE_PATTERN);
		mouseDragRectangle.setMouseTransparent(true);
		// viewer pane
		viewerPane = new Pane();
		layers.boundsInLocalProperty().addListener((obs, ov, nv) -> {
			updateScrollBars();
		});
		layers.getTransforms().add(transform);

		viewerPane.getChildren().add(layers);
		viewerPane.widthProperty().addListener(evt -> {
			updateGridCanvasSize();
			updateScrollBars();
		});
		viewerPane.heightProperty().addListener(evt -> {
			updateGridCanvasSize();
			updateScrollBars();
		});
		// focus event handling
		viewerPane.focusedProperty().addListener((obs, ov, nv) -> {
			viewerManager.handleFocus(nv, getLayers());
		});

		viewerPane.addEventFilter(MouseEvent.MOUSE_CLICKED, evt -> {
			if(evt.isAltDown()) {
				Point2D p = layers.sceneToLocal(evt.getSceneX(), evt.getSceneY());
				scrollToCenter(new DiagramCoordinate(p.getX(), p.getY()));
			}
		});

		// grid pane for viewer and scroll bars
		scrolledPane = new GridPane();
		scrolledPane.add(viewerPane, 1, 1);
		setHgrow(viewerPane, ALWAYS);
		setVgrow(viewerPane, ALWAYS);

		verticalScrollbar = createContentScrollBar(VERTICAL);
		scrolledPane.add(verticalScrollbar, 2, 1);

		horizontalScrollbar = createContentScrollBar(HORIZONTAL);
		scrolledPane.add(horizontalScrollbar, 1, 2);

		viewerPane.getChildren().add(debugLabel);
		debugLabel.setLayoutX(20);
		debugLabel.setLayoutY(20);
		debugLabel.setTextFill(ORANGERED);

		viewerPane.getChildren().add(helpLabel);
		Image helpImage = getFXImage(ToolingCommonUIActivator.PLUGIN_ID, "/icons/help.png");
		ImageView view = new ImageView(helpImage);
		helpLabel.setGraphic(view);
		helpLabel.setLayoutX(viewerPane.getWidth() - helpLabel.getWidth());
		helpLabel.setLayoutY(0);
		helpLabel.setOnMouseClicked(evt -> {
			toggleHelp();
		});

		// clipping rectangle for the viewer pane
		Rectangle viewerClip = new Rectangle();
		viewerPane.setClip(viewerClip);
		viewerPane.layoutBoundsProperty().addListener((obs, ov, nv) -> {
			double width = nv.getWidth();
			viewerClip.setWidth(width);
			viewerClip.setHeight(nv.getHeight());
			helpLabel.setLayoutX(width - helpLabel.getWidth());
		});

		// update the viewer content
		viewerManager.updateContentObjects();

		// add event filters for scrolling and moving viewerPane
		viewerPane.addEventFilter(ScrollEvent.SCROLL, getScrollingFilter());
		ContentDragController contentDragController = new ContentDragController();
		contentDragController.registerEventFilters(viewerPane);
	}

	/** Constructs a scrollbar controlling the content */
	private ScrollBar createContentScrollBar(Orientation orientation) {
		ScrollBar scrollbar = new ScrollBar();
		scrollbar.setOrientation(orientation);
		scrollbar.valueProperty().addListener((obs, ov, nv) -> {
			hideContextMenu();
			Point2D diagramTranslation = layers.parentToLocal(0, 0);
			switch(orientation) {
				case HORIZONTAL:
					transform.appendTranslation(-nv.doubleValue() + diagramTranslation.getX(), 0);
					break;
				case VERTICAL:
					transform.appendTranslation(0, -nv.doubleValue() + diagramTranslation.getY());
					break;
			}
			viewerManager.gridCanvasVisual.updateNodes(layers);
		});
		return scrollbar;
	}

	/** Returns viewerPane. */
	/* package */ Pane getViewerPane() {
		return viewerPane;
	}

	/** Returns the {@link IDiagramMVCBundle}. */
	/* package */ IDiagramMVCBundle getDiagramBundle() {
		return viewerManager.diagramBundle;
	}

	/** Shows the context menu when the mouse is pressed. */
	/* package */ void hideContextMenu() {
		if(contextMenu != null && contextMenu.isShowing()) {
			contextMenu.hide();
			contextMenu = null;
		}
	}

	/** Shows the context menu when the mouse is released. */
	public void showContextMenu(Node node, DiagramCoordinate diagramLocation, IMVCBundle mvcb) {
		List<MenuItem> items = mvcb.getController().contextMenuContributions(node, diagramLocation);
		if(items == null || items.isEmpty()) {
			return;
		}
		contextMenu = new ContextMenu();
		contextMenu.getItems().addAll(items);
		contextMenu.setAutoHide(true);
		Point2D screenLocation = layers.localToScreen(diagramLocation);
		contextMenu.show(node, screenLocation.getX(), screenLocation.getY());
	}

	/** Returns the root model element displayed by this diagram viewer. */
	public Object getRootElement() {
		return viewerManager.getRootModel();
	}

	/** Returns the features instance. */
	public DiagramViewerFeatures getFeatures() {
		return features;
	}

	/**
	 * Returns the content of a SVG representation of this viewers content.
	 * <P>
	 * Note that the background canvas is not exported.
	 */
	public String getSVGExport() {
		SVGExporter exporter = new SVGExporter(this);
		return exporter.export();
	}

	/** Updates the scrollbars. */
	private void updateScrollBars() {
		// The size of the content (in diagram coordinates)
		Bounds contentBounds = layers.getBoundsInLocal();
		double contentHeight = contentBounds.getHeight();
		double contentWidth = contentBounds.getWidth();
		// The size of the visible content (in diagram coordinates)
		Bounds visibleBounds = layers.parentToLocal(viewerPane.getLayoutBounds());
		double visibleWidth = visibleBounds.getWidth();
		double visibleHeight = visibleBounds.getHeight();

		verticalScrollbar.setMin(0);
		horizontalScrollbar.setMin(0);
		verticalScrollbar.setMax(contentHeight);
		horizontalScrollbar.setMax(contentWidth);

		Point2D diagramTopLeft = layers.parentToLocal(0, 0);
		verticalScrollbar.setValue(diagramTopLeft.getY());
		horizontalScrollbar.setValue(diagramTopLeft.getX());

		// Including blank space to the bottom, the height of the content
		// is (contentHeight + visibleHeight), of which visibleHeight is
		// visible. Since the vertical scrollbar goes from 0 to contentHeight,
		// this means we need:
		// contentHeight / verticalVisibleAmount
		// == (contentHeight + visibleHeight) / visibleHeight
		// Hence:
		double verticalVisibleAmount =
				contentHeight * visibleHeight / (contentHeight + visibleHeight);
		double horizontalVisibleAmount =
				contentWidth * visibleWidth / (contentWidth + visibleWidth);
		verticalScrollbar.setVisibleAmount(verticalVisibleAmount);
		horizontalScrollbar.setVisibleAmount(horizontalVisibleAmount);

		verticalScrollbar.setUnitIncrement(1);
		horizontalScrollbar.setUnitIncrement(1);
		verticalScrollbar.setBlockIncrement(verticalScrollbar.getVisibleAmount());
		horizontalScrollbar.setBlockIncrement(horizontalScrollbar.getVisibleAmount());
	}

	/** Update the size of the grid canvas. */
	/* package */ void updateGridCanvasSize() {
		Bounds viewer = viewerPane.getBoundsInLocal();
		viewerManager.gridCanvasVisual.setWidth(viewer.getWidth());
		viewerManager.gridCanvasVisual.setHeight(viewer.getHeight());
	}

	/** Returns the viewer's visual node. */
	public Pane getVisualNode() {
		return scrolledPane;
	}

	/** Returns the current horizontal scroll bar value. */
	public double getHorizontalScrollBarValue() {
		return horizontalScrollbar.getValue();
	}

	/** Returns the current vertical scroll bar value. */
	public double getVerticalScrollBarValue() {
		return verticalScrollbar.getValue();
	}

	/** Returns the model modifier. */
	public Consumer<Change> getModelModifier() {
		return viewerManager.getModelModifier();
	}

	/**
	 * Calls the delete method of all controllers of currently selected elements and returns a
	 * corresponding {@link Change}.
	 */
	public Change deleteAllSelected() {
		final DiagramViewerSelection sel = getSelection();
		return () -> {
			sel.getPrimarySelection().getController().delete();
			for(IMVCBundle sec : sel.getSecondarySelections()) {
				sec.getController().delete();
			}
		};
	}

	/** Returns the current selection. */
	public DiagramViewerSelection getSelection() {
		return viewerManager.getSelection();
	}

	/** Sets the selected {@link IMVCBundle} as single primary selection. */
	public void setSingleSelectedMVCBundle(IMVCBundle sel) {
		viewerManager.setSingleSelectedMVCBundle(sel);
	}

	/** Returns whether this bundle is selected as primary selection. */
	public boolean isPrimarySelected(IMVCBundle bundle) {
		return viewerManager.isPrimarySelected(bundle);
	}

	/** Returns whether this bundle is selected as secondary selection. */
	public boolean isSecondarySelected(IMVCBundle bundle) {
		return viewerManager.isSecondarySelected(bundle);
	}

	/**
	 * Returns whether this bundle is selected (either primary or secondary
	 * selection).
	 */
	public boolean isSelected(IMVCBundle bundle) {
		return isPrimarySelected(bundle) || isSecondarySelected(bundle);
	}

	/**
	 * Adds the given bundle as primary selection and moves the current selection to
	 * the secondary selections list.
	 */
	public void addSelectedMVCBundle(IMVCBundle sel) {
		viewerManager.addSelectedMVCBundle(sel);
	}

	/** Selects all content, diagram anchorage and link {@link IMVCBundle}s of this viewer. */
	public void selectAll() {
		viewerManager.selectAll();
	}

	/** Interprets the effect of shift-clicking the given {@link IMVCBundle}. */
	public void handleCtrlSelectionOf(IMVCBundle sel) {
		viewerManager.handleCtrlSelectionOf(sel);
	}

	/**
	 * Removes the given bundle from either the primary selection (selecting next
	 * the secondary list element) or from the secondary selection list.
	 */
	public void removeSelectedMVCBundle(IMVCBundle sel) {
		viewerManager.removeSelectedMVCBundle(sel);
	}

	/** Returns the visual factory. */
	public IVisualFactory getVisualFactory() {
		return viewerManager.visualFactory;
	}

	/** Returns the controller factory. */
	public IControllerFactory getControllerFactory() {
		return viewerManager.controllerFactory;
	}

	/** Returns the model factory. */
	public IModelFactory getModelFactory() {
		return viewerManager.modelFactory;
	}

	/** Clamp content so that it can't be moved beyond the borders. */
	private void enforceBounds() {
		// The size of the content (in diagram coordinates)
		Bounds contentBounds = layers.getBoundsInLocal();
		// Crop at the top left corner
		BoundingBox enforceContentBounds =
				new BoundingBox(0, 0, contentBounds.getMaxX(), contentBounds.getMaxY());
		Bounds contentInParent = layers.localToParent(enforceContentBounds);

		double leftBoundCorrection = max(contentInParent.getMinX(), 0);
		double topBoundCorrection = max(contentInParent.getMinY(), 0);
		double rightBoundCorrection = min(contentInParent.getMaxX(), 0);
		double bottomBoundCorrection = min(contentInParent.getMaxY(), 0);
		Point2D clampedInParent = new Point2D(leftBoundCorrection + rightBoundCorrection,
				topBoundCorrection + bottomBoundCorrection);
		Point2D correctionLocal = layers.parentToLocal(clampedInParent);
		Point2D originLocal = layers.parentToLocal(0, 0);
		transform.appendTranslation(originLocal.getX() - correctionLocal.getX(),
				originLocal.getY() - correctionLocal.getY());
	}

	/** Apply a transformation to the content. */
	void appendContentTransform(Transform t) {
		transform.append(t);
		enforceBounds();
		updateScrollBars();
		viewerManager.gridCanvasVisual.updateNodes(layers);
	}

	/** Returns the scroll event handler. */
	private EventHandler<? super ScrollEvent> getScrollingFilter() {
		return evt -> {
			evt.consume();
			if(evt.isControlDown()) {
				if(features.getZoomFactorIndex() == -1) {
					return;
				}
				if(evt.getDeltaY() > 0) {
					Point2D pivot = layers.sceneToLocal(evt.getSceneX(), evt.getSceneY());
					features.zoomIn(pivot.getX(), pivot.getY());
				} else if(evt.getDeltaY() < 0) {
					Point2D pivot = layers.sceneToLocal(evt.getSceneX(), evt.getSceneY());
					features.zoomOut(pivot.getX(), pivot.getY());
				}
			} else if(evt.isShiftDown()) {
				Point2D p0 = layers.screenToLocal(0, 0);
				Point2D p1 = layers.screenToLocal(1, 1);
				double hmax = horizontalScrollbar.getMax();
				double hmin = horizontalScrollbar.getMin();
				double nval = horizontalScrollbar.getValue() -
						evt.getDeltaY() * abs(p1.getX() - p0.getX());
				horizontalScrollbar.setValue(max(hmin, min(hmax, nval)));
			} else {
				Point2D p0 = layers.screenToLocal(0, 0);
				Point2D p1 = layers.screenToLocal(1, 1);
				double vmax = verticalScrollbar.getMax();
				double vmin = verticalScrollbar.getMin();
				double vval =
						verticalScrollbar.getValue() - evt.getDeltaY() * abs(p1.getY() - p0.getY());
				double hmax = horizontalScrollbar.getMax();
				double hmin = horizontalScrollbar.getMin();
				double hval = horizontalScrollbar.getValue() -
						evt.getDeltaX() * abs(p1.getX() - p0.getX());
				verticalScrollbar.setValue(max(vmin, min(vmax, vval)));
				horizontalScrollbar.setValue(max(hmin, min(hmax, hval)));
			}
		};
	}

	/** Updates every visual (and only its appearance, not the structure). */
	public void updateAllVisuals() {
		viewerManager.updateAllVisuals();
	}

	/**
	 * Updates the nodes selectively adhering to the relation given by the
	 * {@link IMVCBundle}s.
	 */
	public void updateVisual(IMVCBundle bundle) {
		viewerManager.updateVisual(bundle);
	}

	/**
	 * Updates the nodes selectively adhering to the relation given by the
	 * {@link IMVCBundle}s.
	 */
	public void updateModelVisual(Object model) {
		viewerManager.updateModelVisual(model);
	}

	/**
	 * Updates the content of the viewer creating, deleting and updating any visuals
	 * using the underlying model.
	 */
	public void updateFromModel() {
		viewerManager.updateContentObjects();
	}

	/**
	 * Sets the focus to the viewer pane in order to capture key events. This method
	 * must be called by the container when it receives the focus, e.g. via
	 * {@code @Focus} annotation.
	 */
	public void requestFocus() {
		viewerPane.requestFocus();
	}

	/** Returns whether the diagram viewer has the focus or not. */
	public boolean hasFocus() {
		return viewerPane.isFocused();
	}

	/** Sets the debug message. */
	public void setDebugMessage(String msg) {
		debugLabel.setText(msg);
	}

	/** Returns the layer group. */
	public DiagramLayers getLayers() {
		return layers;
	}

	/** Stars the feedback for selection rectangle. */
	public void startSelectionFeedback(Node node, DiagramCoordinate locationInDiagram) {
		if(node == null || locationInDiagram == null) {
			return;
		}
		mouseDragRectangleStartLocation = locationInDiagram;
		mouseDragRectangle.setX(mouseDragRectangleStartLocation.getX());
		mouseDragRectangle.setY(mouseDragRectangleStartLocation.getY());
		mouseDragRectangle.setWidth(0);
		mouseDragRectangle.setHeight(0);
		if(mouseDragRectangle.getParent() == null) {
			layers.getVisualFeedbackLayer().add(mouseDragRectangle, viewerManager.diagramBundle);
		}
		setSingleSelectedMVCBundle(null);
	}

	/** Converts the given absolute diagram coordinates to node-local coordinates. */
	/* package */ DiagramCoordinate convertEventToDiagramCoordinates(Point2D absoluteLocation,
			Node node) {
		return convertEventToDiagramCoordinates(absoluteLocation.getX(), absoluteLocation.getY(),
				node);
	}

	/** Converts the given mouse event coordinates to node-local coordinates. */
	/* package */ DiagramCoordinate convertEventToDiagramCoordinates(double evtX, double evtY,
			Node node) {
		Point2D scene = node.localToScene(evtX, evtY);
		Point2D diagram = layers.sceneToLocal(scene);
		return new DiagramCoordinate(diagram.getX(), diagram.getY());
	}

	/** By default, zoom fixes the top left corner */
	/* package */ DiagramCoordinate getDefaultZoomPivot() {
		Point2D pivot = layers.parentToLocal(0, 0);
		return new DiagramCoordinate(pivot.getX(), pivot.getY());
	}

	/** Scrolls the diagram to center at the given coordinates. */
	public void scrollToCenter(DiagramCoordinate center) {
		Point2D centerScene =
				viewerPane.localToScene(viewerPane.getWidth() / 2, viewerPane.getHeight() / 2);
		Point2D currentCenter = layers.sceneToLocal(centerScene);
		double dx = currentCenter.getX() - center.getX();
		double dy = currentCenter.getY() - center.getY();
		appendContentTransform(new Translate(dx, dy));
	}

	/** Updates the feedback for selection rectangle. */
	public void updateSelectionFeedback(Node node, Point2D locationInDiagram) {
		if(node == null || locationInDiagram == null) {
			return;
		}
		Point2D location = locationInDiagram;
		double x = min(mouseDragRectangleStartLocation.getX(), location.getX());
		mouseDragRectangle.setX(x);
		double w = abs(mouseDragRectangleStartLocation.getX() - location.getX());
		mouseDragRectangle.setWidth(w);
		double y = min(mouseDragRectangleStartLocation.getY(), location.getY());
		mouseDragRectangle.setY(y);
		double h = abs(mouseDragRectangleStartLocation.getY() - location.getY());
		mouseDragRectangle.setHeight(h);
	}

	/** Terminates the feedback for selection rectangle. */
	public void terminateSelectionFeedback(Node node, Point2D locationInDiagram) {
		if(node == null || locationInDiagram == null) {
			return;
		}
		Point2D location = locationInDiagram;
		double x = min(mouseDragRectangleStartLocation.getX(), location.getX());
		double w = abs(mouseDragRectangleStartLocation.getX() - location.getX());
		double y = min(mouseDragRectangleStartLocation.getY(), location.getY());
		double h = abs(mouseDragRectangleStartLocation.getY() - location.getY());
		layers.getVisualFeedbackLayer().remove(mouseDragRectangle);
		Rectangle2D selectionRect = new Rectangle2D(x, y, w, h);
		// remove old selection
		setSingleSelectedMVCBundle(null);
		for(IContentMVCBundle cmvcb : viewerManager.contentBundles.values()) {
			Rectangle2D contentRect = cmvcb.getVisual().getCurrentBounds();
			if(contentRect != null && selectionRect.contains(contentRect)) {
				addSelectedMVCBundle(cmvcb);
			}
		}
		for(IDiagramAnchorageMVCBundle damvcb : viewerManager.diagramAnchorageBundles.values()) {
			Rectangle2D daRect = damvcb.getVisual().getCurrentBounds();
			if(daRect != null && selectionRect.contains(daRect)) {
				addSelectedMVCBundle(damvcb);
			}
		}
	}

	/** Starts the feedback of the link creation line and shows the possible target feedbacks. */
	public void startNewLinkLineFeedback(IMVCBundle linkStartBundle, Node node,
			Point2D locationInDiagram) {
		if(node == null || locationInDiagram == null) {
			return;
		}
		Point2D location = locationInDiagram;
		linkLineFeedback.setStartX(location.getX());
		linkLineFeedback.setStartY(location.getY());
		linkLineFeedback.setEndX(location.getX());
		linkLineFeedback.setEndY(location.getY());
		if(linkLineFeedback.getParent() == null) {
			layers.getVisualFeedbackLayer().add(linkLineFeedback, viewerManager.diagramBundle);
		}
		for(IMVCBundle possibleTarget : viewerManager.possibleLinkTargets) {
			IController controller = possibleTarget.getController();
			MVCBundleTag effect = controller.getLinkTargetEffectForNewLinkFrom(linkStartBundle);
			if(effect != null) {
				possibleTarget.addTag(effect);
				possibleTarget.getVisual().updateNodes(layers);
			}
		}
	}

	/** Updates the end of the link creation line. */
	public void updateNewLinkLineFeedback(Point2D locationInDiagram, Paint color) {
		linkLineFeedback.setEndX(locationInDiagram.getX());
		linkLineFeedback.setEndY(locationInDiagram.getY());
		linkLineFeedback.setStroke(color);
	}

	/** Terminates the link creation line and possible target feedbacks. */
	public void terminateNewLinkLineFeedback() {
		layers.getVisualFeedbackLayer().remove(linkLineFeedback);
		clearLinkTargetFeedback();
	}

	/** Clears the link target feedback from all possible link targets. */
	private void clearLinkTargetFeedback() {
		for(IMVCBundle possibleTarget : viewerManager.possibleLinkTargets) {
			possibleTarget.removeTag(LINK_TARGET_ALLOWED_TAG);
			possibleTarget.removeTag(LINK_TARGET_DENIED_TAG);
			possibleTarget.removeTag(LINK_TARGET_MAYBE_TAG);
			possibleTarget.getVisual().updateNodes(layers);
		}
	}

	/** Starts the feedback for relocating a link. */
	public void startMoveLinkFeedback(ILinkMVCBundle linkBundle,
			IAnchorageMVCBundle movedStartOrEnd) {
		for(IMVCBundle possibleTarget : viewerManager.possibleLinkTargets) {
			IController controller = possibleTarget.getController();
			MVCBundleTag effect =
					controller.getLinkTargetEffectForMoveLink(linkBundle, movedStartOrEnd);
			if(effect != null) {
				possibleTarget.addTag(effect);
				possibleTarget.getVisual().updateNodes(layers);
			}
		}
	}

	/** Terminates the link relocation feedbacks. */
	public void terminateMoveLinkFeedback() {
		clearLinkTargetFeedback();
	}

	/** Returns the last known mouse location. */
	public DiagramCoordinate getLastMouseLocation() {
		return getLayers().getMouseState().getLastMouseLocation();
	}

	/** Returns the current drag extent. */
	public DiagramCoordinate getDragExtent() {
		return getLayers().getMouseState().getDragExtent();
	}

	/** Toggles the in-place help. */
	private void toggleHelp() {
		if(helpText == null) {
			// enable help
			String txt = features.getHelpText();
			if(txt == null || "".equals(txt)) {
				txt = "No in-place help available.";
			}
			helpText = new Text(txt);
			helpText.setFill(Color.WHITE);
			helpText.setLayoutY(0);
			double lw = viewerPane.getWidth() - helpLabel.getWidth();
			// 20% of the width with 3 pixel inset
			helpText.setLayoutX(lw * .2 + 3);
			helpText.setLayoutY(3);
			helpText.setTextOrigin(TOP);
			helpText.setWrappingWidth(lw * .8 - 6);
			helpText.setOnMouseClicked(evt -> toggleHelp());
			helpShadow.setX(lw * .2);
			helpShadow.setY(0);
			helpShadow.setWidth(lw * .8);
			helpShadow.setHeight(viewerPane.getHeight());
			helpShadow.setFill(grayRgb(32, .66));
			helpShadow.setOnMouseClicked(evt -> toggleHelp());

			viewerPane.getChildren().add(helpShadow);
			viewerPane.getChildren().add(helpText);
		} else {
			// disable help
			viewerPane.getChildren().remove(helpText);
			viewerPane.getChildren().remove(helpShadow);
			helpText = null;
		}
	}

	/**
	 * Creates a drag controller for the given bundle, gesture, mouse location and source node also
	 * considering the current selection.
	 */
	public IDragController createDragController(IMVCBundle bundle, EDragGesture gesture,
			Node source, DiagramCoordinate lastMouseLocation) {
		DiagramViewerSelection sel = getSelection();
		if(!sel.isMultiSelection()) {
			// single selection delegates to the underlying controller
			return bundle.getController().getDragController(gesture, source, lastMouseLocation);
		}
		// return a composite controller
		return new CompositeDragController(sel, gesture, source, lastMouseLocation);
	}

	/**
	 * Creates a key press controller for the given bundle also considering the current selection.
	 */
	public IKeyPressController createKeyPressController(IMVCBundle bundle, Node source,
			DiagramCoordinate lastMouseLocation) {
		final DiagramViewerSelection sel = getSelection();
		if(sel.isMultiSelection()) {
			return new IKeyPressController() {
				/** {@inheritDoc} */
				@Override
				public Change keyEvent(KeyEvent event, IMVCBundle bundle, Node source,
						DiagramCoordinate mouseLocation) {
					ChangeSet changeSet = new ChangeSet();
					// handle primary selection
					IMVCBundle primaryBundle = sel.getPrimarySelection();
					IKeyPressController primaryKPC = primaryBundle.getController()
							.getKeyPressController(source, lastMouseLocation);
					Change primaryChg =
							primaryKPC.keyEvent(event, primaryBundle, source, lastMouseLocation);
					if(primaryChg != null) {
						changeSet.add(primaryChg);
					}
					// handle secondary selection
					for(IMVCBundle secondaryBundle : sel.getSecondarySelections()) {
						IKeyPressController secondaryKPC = secondaryBundle.getController()
								.getKeyPressController(source, lastMouseLocation);
						Change secondaryChg = secondaryKPC.keyEvent(event, primaryBundle, source,
								lastMouseLocation);
						if(secondaryChg != null) {
							changeSet.add(secondaryChg);
						}
					}
					if(!changeSet.isEmpty()) {
						return changeSet;
					}
					return null;
				}
			};
		}
		return bundle.getController().getKeyPressController(source, lastMouseLocation);
	}

	/** Composite drag controller delegates to a set of {@link IDragController}s. */
	private static final class CompositeDragController implements IDragController {
		/** The contained controllers. */
		private final Map<IDragController, IMVCBundle> controllers = new HashMap<>();

		/** Constructor. */
		public CompositeDragController(DiagramViewerSelection selection, EDragGesture gesture,
				Node source, DiagramCoordinate lastMouseLocation) {
			addDragController(selection.getPrimarySelection(), gesture, source, lastMouseLocation);
			for(IMVCBundle sec : selection.getSecondarySelections()) {
				addDragController(sec, gesture, source, lastMouseLocation);
			}
		}

		/** Creates a drag controller from the bundle's controller and adds it. */
		private void addDragController(IMVCBundle bundle, EDragGesture gesture, Node source,
				DiagramCoordinate lastMouseLocation) {
			if(bundle == null || bundle.getController() == null) {
				return;
			}
			IDragController dc =
					bundle.getController().getDragController(gesture, source, lastMouseLocation);
			if(dc != null) {
				controllers.put(dc, bundle);
			}
		}

		/** {@inheritDoc} */
		@Override
		public void dragStarted(IMVCBundle startBundle, Node startNode,
				DiagramCoordinate lastMouseLocation) {
			for(IDragController dc : controllers.keySet()) {
				IMVCBundle bundle = controllers.get(dc);
				dc.dragStarted(bundle, startNode, lastMouseLocation);
			}
		}

		/** {@inheritDoc} */
		@Override
		public void dragInProgress(IMVCBundle currentBundle, Node currentNode,
				DiagramCoordinate lastMouseLocation) {
			for(IDragController dc : controllers.keySet()) {
				IMVCBundle bundle = controllers.get(dc);
				dc.dragInProgress(bundle, currentNode, lastMouseLocation);
			}
		}

		/** {@inheritDoc} */
		@Override
		public Change dragCompleted(IMVCBundle endBundle, Node endNode,
				DiagramCoordinate lastMouseLocation) {
			ChangeSet cset = new ChangeSet();
			for(IDragController dc : controllers.keySet()) {
				Change chg = dc.dragCompleted(endBundle, endNode, lastMouseLocation);
				if(chg != null) {
					cset.add(chg);
				}
			}
			if(!cset.isEmpty()) {
				return cset;
			}
			return null;
		}
	}

	/** Mouse controller for moving the content using control-drag. */
	private final class ContentDragController {
		/** Last point within a control-drag gesture. */
		private Point2D dragPoint;
		/**
		 * Has the mouse button been released within a control-drag gesture? If so, the next click
		 * event should be filtered out.
		 */
		private boolean buttonReleased;

		/** Register the event filters of this controller. */
		/* package */ void registerEventFilters(Node node) {
			node.addEventFilter(MouseEvent.MOUSE_DRAGGED, dragDetected);
			node.addEventFilter(MouseEvent.ANY, dragged);
			node.addEventFilter(MouseEvent.MOUSE_RELEASED, mouseReleased);
			node.addEventFilter(MouseEvent.MOUSE_CLICKED, mouseClicked);
		}

		/** Control-drag gesture starts when drag is detected. */
		private final EventHandler<MouseEvent> dragDetected = evt -> {
			if(dragPoint == null && evt.isControlDown() && evt.getButton() == MouseButton.PRIMARY) {
				dragPoint = getLayers().sceneToLocal(evt.getSceneX(), evt.getSceneY());
				buttonReleased = false;
				evt.consume();
			}
		};

		/** While dragging, the content is translated appropriately. */
		private final EventHandler<MouseEvent> dragged = evt -> {
			if(dragPoint != null) {
				evt.consume();

				Point2D evtLayer = getLayers().sceneToLocal(evt.getSceneX(), evt.getSceneY());

				Point2D move = evtLayer.subtract(dragPoint);
				double dx = move.getX();
				double dy = move.getY();
				appendContentTransform(new Translate(dx, dy));
				dragPoint = getLayers().sceneToLocal(evt.getSceneX(), evt.getSceneY());
			}
		};

		/** If the mouse button is released, the gesture should be stopped. */
		private final EventHandler<MouseEvent> mouseReleased = evt -> {
			if(dragPoint != null && evt.getButton() == MouseButton.PRIMARY) {
				dragPoint = null;
				buttonReleased = true;
				evt.consume();
			}
		};

		/** The first click event after the released event should be filtered. */
		private final EventHandler<MouseEvent> mouseClicked = evt -> {
			if(buttonReleased && evt.getButton() == MouseButton.PRIMARY) {
				evt.consume();
			}
			buttonReleased = false;
		};
	}
}
