/*******************************************************************************
 * Copyright (c) 2019, 2020 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Apache License, Version 2.0 which is available at
 * https://www.apache.org/licenses/LICENSE-2.0.
 *
 * SPDX-License-Identifier: Apache-2.0
 *******************************************************************************/
package org.fortiss.tooling.common.ui.javafx.lwfxef;

import static java.lang.Double.doubleToLongBits;

import java.util.Objects;

import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramViewerFeatures.IndicatorType;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.visual.IVisual;

import javafx.geometry.Bounds;
import javafx.geometry.Point2D;
import javafx.geometry.Rectangle2D;
import javafx.scene.Node;
import javafx.scene.canvas.Canvas;
import javafx.scene.canvas.GraphicsContext;
import javafx.scene.layout.Pane;
import javafx.scene.paint.Color;
import javafx.scene.shape.Rectangle;

/**
 * {@link IVisual} for the diagram background. Note that the {@link Canvas} used here is always as
 * big as the diagram viewer's content area and the grid is painted with an according displacement.
 * Having the canvas as big as the diagram will most certainly result in video memory problems,
 * which requires to implement the background in this special way.
 */
final class GridCanvasVisual implements IVisual {
	/** The diagram bundle. */
	private final DiagramViewer viewer;

	/** The pane to contain the whole grid. */
	private final Pane pane;

	/**
	 * The FX canvas to draw the grid. The canvas is one grid column/row larger than the pane. It
	 * gets moved on the pane when the position of the grid changes. In this way, the canvas only
	 * needs to be redrawn when scale or grid configuration change. Moving the final image is much
	 * faster than redrawing the canvas.
	 */
	private final Canvas gridCanvas;

	/**
	 * A rectangle for the left outer border. The border is not painted with the canvas to avoid
	 * having to redraw the canvas when changing location.
	 */
	private final Rectangle leftOuterBorder;

	/** A rectangle for the top outer border. */
	private final Rectangle topOuterBorder;

	/** Width of the whole shown grid. */
	double width;

	/** Height of the whole shown grid. */
	double height;

	/** Constructor. */
	public GridCanvasVisual(DiagramViewer viewer) {
		this.viewer = viewer;
		this.pane = new Pane();
		this.gridCanvas = new Canvas();
		this.leftOuterBorder = new Rectangle();
		this.topOuterBorder = new Rectangle();
		this.pane.getChildren().addAll(gridCanvas, leftOuterBorder, topOuterBorder);
	}

	/** Returns gridCanvas. */
	/* package */Node getGridNode() {
		return pane;
	}

	/** {@inheritDoc} */
	@Override
	public boolean enableInteraction() {
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public boolean enableVisual() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public IMVCBundle getMVCBundle() {
		return viewer.getDiagramBundle();
	}

	/** {@inheritDoc} */
	@Override
	public void updateNodes(DiagramLayers layers) {
		if(pane.getParent() == null) {
			// grid is added below the diagram layers
			viewer.getViewerPane().getChildren().add(0, pane);
			viewer.getLayers().getMouseState().registerMouseListeners(pane, getMVCBundle());
		}
		updateContent();
	}

	/** {@inheritDoc} */
	@Override
	public void removeAllVisuals(DiagramLayers layers) {
		viewer.getLayers().getMouseState().unregisterMouseListeners(pane);
		viewer.getViewerPane().getChildren().remove(pane);
	}

	/** {@inheritDoc} */
	@Override
	public Rectangle2D getModelBounds() {
		Bounds b = pane.getBoundsInParent();
		return new Rectangle2D(b.getMinX(), b.getMinY(), b.getWidth(), b.getHeight());
	}

	/** {@inheritDoc} */
	@Override
	public DiagramCoordinate getLinkAnchorage(DiagramCoordinate indication) {
		DiagramViewerFeatures features = viewer.getFeatures();
		return indication.add(features.getHorizontalSpacing() / 2,
				features.getVerticalSpacing() / 2);
	}

	/** {@inheritDoc} */
	@Override
	public Rectangle2D getCurrentBounds() {
		return getModelBounds();
	}

	/** {@inheritDoc} */
	@Override
	public void requestFocus() {
		pane.requestFocus();
	}

	/** Returns the zoom factor from current content transformation. */
	private double getZoomFactor() {
		Point2D unitVector = new Point2D(1, 1);
		Point2D transformedVector =
				viewer.getLayers().getLocalToParentTransform().deltaTransform(unitVector);
		return transformedVector.getX();
	}

	/** Structure to represent the current location of the grid. */
	private static class GridOffset {
		/** Grid x coordinate of upper left corner. */
		int gridX;

		/** Grid y coordinate of upper left corner. */
		int gridY;

		/** Insets in the grid square between (gridX, gridY) and (gridX + 1, gridY +1). */
		Point2D insets;
	}

	/** Compute the grid offset current content transformation. */
	private GridOffset getGridOffset() {
		DiagramViewerFeatures features = viewer.getFeatures();
		Point2D p0 = viewer.getLayers().sceneToLocal(pane.localToScene(0, 0));

		GridOffset gridOffset = new GridOffset();
		gridOffset.gridX = (int)(p0.getX() / features.getHorizontalSpacing());
		gridOffset.gridY = (int)(p0.getY() / features.getVerticalSpacing());
		gridOffset.insets = new Point2D((p0.getX() / features.getHorizontalSpacing()) % 1.0,
				(p0.getY() / features.getVerticalSpacing()) % 1.0);
		return gridOffset;
	}

	/**
	 * Structure representing the parameters that go into the drawing of the grid canvas. It is
	 * useful to determine if the canvas needs repainting.
	 */
	private static class GridParameters {

		/** Horizontal grid spacing. */
		private double hSpacing;

		/** Vertical grid spacing. */
		private double vSpacing;

		/** Zoom factor. */
		private double zoomFactor;

		/** Grid indicator size. */
		private double indicatorSize;

		/** Grid indicator type. */
		private IndicatorType indicatorType;

		/** Grid background color. */
		private Color backgroundColor;

		/** Grid indicator color. */
		private Color indicatorColor;

		/** {@inheritDoc} */
		@Override
		public int hashCode() {
			return Objects.hash(backgroundColor, hSpacing, indicatorColor, indicatorSize,
					indicatorType, vSpacing, zoomFactor);
		}

		/**
		 * Auto-generated (by Eclipse) deep equality method.
		 * 
		 * {@inheritDoc}
		 */
		@Override
		public boolean equals(Object obj) {
			if(this == obj) {
				return true;
			}
			if(obj == null) {
				return false;
			}
			if(getClass() != obj.getClass()) {
				return false;
			}
			GridParameters other = (GridParameters)obj;

			// Compare all member variables individually for equality.
			// The equality of doubles is tested using {@code doubleToLongBits},
			// as explained in {@link java.lang.Double#equals(Object)}.
			return Objects.equals(backgroundColor, other.backgroundColor) &&
					doubleToLongBits(hSpacing) == doubleToLongBits(other.hSpacing) &&
					Objects.equals(indicatorColor, other.indicatorColor) &&
					doubleToLongBits(indicatorSize) == doubleToLongBits(other.indicatorSize) &&
					indicatorType == other.indicatorType &&
					doubleToLongBits(vSpacing) == doubleToLongBits(other.vSpacing) &&
					doubleToLongBits(zoomFactor) == doubleToLongBits(other.zoomFactor);
		}
	}

	/** Update the pane node. */
	private void updateContent() {
		DiagramViewerFeatures features = viewer.getFeatures();

		GridParameters parameters = new GridParameters();
		parameters.hSpacing = features.getHorizontalSpacing();
		parameters.vSpacing = features.getVerticalSpacing();
		parameters.zoomFactor = getZoomFactor();
		parameters.indicatorType = features.getIndicatorType();
		parameters.backgroundColor = features.getBackgroundColor();
		parameters.indicatorColor = features.getIndicatorColor();
		parameters.indicatorSize = features.getIndicatorSize();

		paintCanvas(parameters);

		GridOffset gridOffset = getGridOffset();
		double zfhs = parameters.zoomFactor * parameters.hSpacing;
		double zfvs = parameters.zoomFactor * parameters.vSpacing;
		gridCanvas.relocate(-gridOffset.insets.getX() * zfhs, -gridOffset.insets.getY() * zfvs);

		pane.getChildren().removeAll(leftOuterBorder, topOuterBorder);
		if(features.isDrawOuterBorder()) {
			pane.getChildren().addAll(leftOuterBorder, topOuterBorder);
			leftOuterBorder.relocate(-(gridOffset.gridX + gridOffset.insets.getX()) * zfhs, 0);
			leftOuterBorder.setWidth(zfhs);
			leftOuterBorder.setHeight(height);
			leftOuterBorder.setStrokeWidth(0.0);
			leftOuterBorder.setFill(features.getIndicatorColor());

			topOuterBorder.relocate(0, -(gridOffset.gridY + gridOffset.insets.getY()) * zfhs);
			topOuterBorder.setWidth(width);
			topOuterBorder.setHeight(zfvs);
			topOuterBorder.setStrokeWidth(0.0);
			topOuterBorder.setFill(features.getIndicatorColor());
		}
	}

	/** The parameters of what is currently drawn on the gridCanvas. */
	private GridParameters currentGridCanvasParameters = null;

	/** Draw gridCanvas with the given parameters. */
	private void paintCanvas(GridParameters parameters) {
		double targetCanvasWidth = width + parameters.zoomFactor * parameters.hSpacing;
		double targetCanvasHeight = height + parameters.zoomFactor * parameters.vSpacing;

		boolean resized = false;
		resized |= gridCanvas.getWidth() != targetCanvasWidth;
		resized |= gridCanvas.getHeight() != targetCanvasHeight;

		if(resized) {
			gridCanvas.setWidth(targetCanvasWidth);
			gridCanvas.setHeight(targetCanvasHeight);
		}

		if(!resized && parameters.equals(currentGridCanvasParameters)) {
			// grid is already drawn on canvas
			return;
		}
		currentGridCanvasParameters = parameters;

		GraphicsContext gc = gridCanvas.getGraphicsContext2D();
		// background
		gc.setFill(parameters.backgroundColor);
		gc.setStroke(parameters.backgroundColor);
		gc.fillRect(0, 0, gridCanvas.getWidth(), gridCanvas.getHeight());

		// abort if no indicators should be drawn
		IndicatorType indicatorType = parameters.indicatorType;
		if(indicatorType == IndicatorType.INVISIBLE) {
			return;
		}

		// draw grid indicators
		gc.setStroke(parameters.indicatorColor);
		gc.setFill(parameters.indicatorColor);
		double zfhs = parameters.zoomFactor * parameters.hSpacing;
		double zfvs = parameters.zoomFactor * parameters.vSpacing;
		double isize = parameters.indicatorSize * parameters.zoomFactor;

		for(double x = zfhs; x < gridCanvas.getWidth(); x += zfhs) {
			for(double y = zfvs; y < gridCanvas.getHeight(); y += zfvs) {
				drawIndicator(gc, Math.floor(x) + 0.5, Math.floor(y) + 0.5, isize, indicatorType);
			}
		}
	}

	/** Draws the indicators. */
	private void drawIndicator(GraphicsContext gc, double x, double y, double size,
			IndicatorType indicatorType) {
		switch(indicatorType) {
			case CIRCLE:
				drawCircle(gc, x, y, size);
				break;
			case CROSS:
				drawCross(gc, x, y, size);
				break;
			case SQUARE:
				drawSquare(gc, x, y, size);
				break;
			default:
				break;
		}
	}

	/** Draws a small cross at the given point. */
	private void drawCircle(GraphicsContext gc, double x, double y, double size) {
		gc.fillOval(x - size, y - size, 2 * size, 2 * size);
	}

	/** Draws a small cross at the given point. */
	private void drawCross(GraphicsContext gc, double x, double y, double size) {
		gc.strokeLine(x - size, y, x + size, y);
		gc.strokeLine(x, y - size, x, y + size);
	}

	/** Draws a small square at the given point. */
	private void drawSquare(GraphicsContext gc, double x, double y, double size) {
		gc.fillRect(x - size, y - size, 2 * size, 2 * size);
	}

	/** {@inheritDoc} */
	@Override
	public EDragGesture getDragGesture(Node node, DiagramCoordinate diagramLocation) {
		if(node == pane) {
			return EDragGesture.SELECTION;
		}
		return EDragGesture.NONE;
	}

	/** Sets the width of the grid canvas. */
	public void setWidth(double w) {
		this.width = w;
		updateContent();
	}

	/** Sets the height of the grid canvas. */
	public void setHeight(double h) {
		this.height = h;
		updateContent();
	}
}
