/*-------------------------------------------------------------------------+
| Copyright 2021 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.extension.exception;

import java.util.List;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.kernel.extension.data.ITransformationContext;
import org.fortiss.tooling.kernel.utils.TransformationUtils;

/**
 * An exception thrown by
 * {@link TransformationUtils#createTransformedObjectFor(EObject, Class, ITransformationContext)}
 * if all chain transformations fail, keeping track of all exceptions thrown by each failed
 * provider.
 * 
 * @author muntean
 */
public class AllChainTransformationsFailedException extends ChainTransformationFailedException {

	/** Number of tabs initially used when returning the exception's message. */
	private final int INITIAL_TABS = 0;

	/** Stores the {@link ChainTransformationFailedException}s thrown by each failed provider. */
	private final List<ChainTransformationFailedException> exceptionList;

	/** Constructor. */
	public AllChainTransformationsFailedException(
			List<ChainTransformationFailedException> exceptionList) {
		super(null, null, null, null);
		this.exceptionList = exceptionList;
	}

	/** {@inheritDoc} */
	@Override
	public String getMessage() {
		return getMessage(INITIAL_TABS);
	}

	/**
	 * Recursive method iterating through the caught {@link TransformationFailedException}s
	 * backtracking to the source of the error by examining the cause of the exceptions.
	 * 
	 * @param counter
	 *            Number of tabs used in the current recursive step.
	 * @return Message expressing the failed provider + cause.
	 */
	private String getMessage(int counter) {
		String msg = "";
		String tabs = addTabs(counter);
		for(TransformationFailedException ctfe : this.exceptionList) {
			msg += tabs + "Failed Provider:\n";
			msg += tabs + ctfe.getFailedProvider().getClass().getSimpleName() + "\n";

			// Here, the cause is either a...
			// ... ChainTransformationsFailedException or an
			// ... AllChainTransformationsFailedException.
			TransformationFailedException cause = (TransformationFailedException)ctfe.getCause();
			if(cause != null) {
				msg += tabs + "Cause:\n";
				if(cause instanceof AllChainTransformationsFailedException)
					msg += ((AllChainTransformationsFailedException)cause).getMessage(counter + 1) +
							"\n";
				else
					msg += tabs + cause.getMessage() + "\n";
			}
		}
		return msg;
	}

	/** Returns a string with as many tabs as specified by the parameter. */
	private String addTabs(int counter) {
		String msg = "";
		for(int i = 0; i < counter; i++) {
			msg += "\t";
		}
		return msg;
	}

	/** Returns the list {@link ChainTransformationFailedException}s. */
	public List<ChainTransformationFailedException> getExceptionList() {
		return exceptionList;
	}
}
